"""
Import this module to have the side effect
that the RETURNN root dir will be added to ``sys.path``,
i.e. ``import returnn`` works afterwards.

In your code (e.g. tools or demos), you would have this::

    import _setup_returnn_env  # noqa

The ``# noqa`` is to ignore the warning that this module is not used.
"""

import os
import sys

my_dir = os.path.dirname(os.path.realpath(os.path.abspath(__file__)))
root_dir = os.path.dirname(my_dir)
sys.path.insert(0, root_dir)



"""
This is here so that the RETURNN repo root directory can be imported as a submodule.
This was used for the old flat code file structure, for usage like::

    import returnn.TFUtil

We want to support the same code.
"""

from __future__ import annotations
import os
import sys

if globals().get("__package__", None) is None:
    __package__ = __name__  # https://www.python.org/dev/peps/pep-0366/ # pylint: disable=redefined-builtin

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

if __name__ == "returnn":
    # This is a bit tricky. `import returnn` in here will not work, as this is the reference to ourselves!
    # Thus delete this ref, to enforce a reload of the right module.
    # For the user, this will actually replace the `returnn` module by the right one.
    # But this should be compatible.
    sys.modules.pop("returnn", None)
    # It is important that we have some `import returnn` below.

from returnn.__old_mod_loader__ import (
    setup as _old_module_loader_setup,
)  # nopep8 # pylint: disable=wrong-import-position

# noinspection PyUnboundLocalVariable
_old_module_loader_setup(package_name=__package__, modules=globals())



"""
tests for returnn.tensor
"""

import numpy
import _setup_test_env  # noqa
import sys
import unittest
from returnn.util import better_exchook
import returnn.frontend as rf
from returnn.tensor import Dim, Tensor, TensorDict


def test_tensor():
    batch_dim = Dim(name="batch", dimension=None)
    time_dim = Dim(name="time", dimension=None)
    feat_dim = Dim(10)
    x = Tensor("x", (batch_dim, time_dim, feat_dim), "float32")
    print(x)


# Note: Some of the dim math tests are duplicated from test_TFUtil.py, and partly adapted.
# But here they are TF-independent.


def test_dim_math_basics():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert a == a
    assert (a + 2 - 2) == a
    assert a + b == a + b
    assert a + b != b + a  # not commutative
    assert a * b == a * b
    assert a * b != b * a  # not commutative
    assert 2 * a == a + a
    assert 3 * a == a + a + a
    assert a * 2 != 2 * a
    assert 2 * a + b == a + a + b
    assert a + b - b == a
    assert a + 2 * b - b + -b == a
    assert a * b + b == (a + 1) * b
    assert (a + b) * 2 == a * 2 + b * 2
    assert 0 + a + 0 == a
    assert sum([0, a, 0, a, 0]) == 2 * a


def test_dim_math_neq_after_inc():
    a = Dim(0, name="a")
    b = Dim(0, name="a")  # same name, but different instance, intentionally to trigger potential issues
    assert a != b
    assert a + 1 != b + 1
    # Note: We had the bug that _representative_tag selected an auto-generated dim
    # (the unnamed "1" dim is auto-generated),
    # and that was used as derived_from_tag, and that triggered that the resulting dim was also auto-generated,
    # and for auto-generated dims, we allow dim equality also by name.
    # The fix was: _representative_tag will always prefer a non-auto-generated dim if there is any.
    assert a + 1 + 1 != b + 1 + 1


def test_dim_math_double_neg():
    a = Dim(None, name="a")
    assert --a == a


def test_dim_math_mul_rmul():
    a = Dim(None, name="a")
    b = a * 3
    c = 2 * b
    assert c == 2 * (a * 3)
    assert c != a * 6
    assert c == (2 * a) * 3


def test_dim_math_mul_div():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert (a * b) // b == a
    assert (b * a) // b != a
    assert (b * a).div_left(b) == a


def test_dim_math_div_mul():
    a = Dim(None, name="a")
    b = Dim(None, name="b")
    assert a // b == a // b


def test_dim_math_div_div():
    a = Dim(None, name="a")
    b = a.ceildiv_right(2)
    b = b.ceildiv_right(3)
    c = a.ceildiv_right(6)
    print(a, b, c)
    assert b == c


def test_dim_math_pad_conv():
    time = Dim(None, name="time")
    padded = 2 + time + 2
    assert padded == 2 + time + 2
    conv_valid = (-2) + padded + (-2)
    assert conv_valid == time


def test_dim_math_pad_window():
    rf.select_backend("numpy")
    batch_dim = Dim(3, name="batch")
    time1_dim = Dim(rf.convert_to_tensor([5, 6, 7], dims=[batch_dim], name="time1"))
    time2_dim = Dim(rf.convert_to_tensor([20, 21, 22], dims=[batch_dim], name="time2"))
    time3_dim = Dim(rf.convert_to_tensor([10, 5, 3], dims=[batch_dim], name="time3"))
    in_spatial_dim = time1_dim + time2_dim + time3_dim
    filter_size = 17
    # As it would happen with conv/pool/window/... with window size 17.
    out_spatial_dim = in_spatial_dim.sub_left(8).sub_right(8)
    print("out_spatial_dim:", out_spatial_dim)
    sizes = out_spatial_dim.dyn_size.tolist()
    print("sizes:", sizes)
    expected_sizes = (time1_dim.dyn_size + time2_dim.dyn_size + time3_dim.dyn_size - (filter_size - 1)).tolist()
    print("expected_sizes:", expected_sizes)
    assert sizes == expected_sizes


def test_Tensor_pickle():
    import pickle

    tensor = Tensor("data", dims=[Dim(10)], dtype="int32", raw_tensor=numpy.zeros((10,), dtype="int32"))

    s = pickle.dumps(tensor)
    tensor2: Tensor = pickle.loads(s)

    assert tensor.name == tensor2.name
    assert tensor.dtype == tensor2.dtype
    assert tensor.raw_tensor is not None
    assert tensor2.raw_tensor is not None
    assert numpy.array_equal(tensor.raw_tensor, tensor2.raw_tensor)


def test_TensorDict_pickle():
    import pickle

    tensor_dict = TensorDict(
        {"data": Tensor("data", dims=[Dim(10)], dtype="int32", raw_tensor=numpy.zeros((10,), dtype="int32"))}
    )
    assert all(tensor.raw_tensor is not None for tensor in tensor_dict.data.values())

    s = pickle.dumps(tensor_dict)
    tensor_dict2 = pickle.loads(s)

    assert tensor_dict.data.keys() == tensor_dict2.data.keys()
    assert all(tensor.raw_tensor is not None for tensor in tensor_dict2.data.values())


def test_TensorDict_queue():
    from multiprocessing import Queue

    tensor_dict = TensorDict(
        {"data": Tensor("data", dims=[Dim(10)], dtype="int32", raw_tensor=numpy.zeros((10,), dtype="int32"))}
    )
    assert all(tensor.raw_tensor is not None for tensor in tensor_dict.data.values())

    q = Queue(maxsize=2)
    q.put(tensor_dict)
    tensor_dict2 = q.get()

    assert tensor_dict.data.keys() == tensor_dict2.data.keys()
    assert all(tensor.raw_tensor is not None for tensor in tensor_dict2.data.values())


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
HuggingFace dataset wrapper

See https://github.com/rwth-i6/returnn/issues/1257 for some initial discussion.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, Any, Callable, Sequence, Dict, List
import os
import re
import numpy
from returnn.tensor import Tensor
from returnn.util import file_cache
from .basic import DatasetSeq
from .cached2 import CachedDataset2
from .util.vocabulary import Vocabulary
from .util.strings import str_to_numpy_array

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import datasets


class HuggingFaceDataset(CachedDataset2):
    """
    HuggingFace dataset wrapper.
    """

    def __init__(
        self,
        dataset_opts: Union[
            Dict[str, Any],
            str,
            os.PathLike,
            Sequence[Union[str, os.PathLike]],
            Callable[[], Union[Dict[str, Any], str, os.PathLike, Sequence[Union[str, os.PathLike]], datasets.Dataset]],
        ],
        *,
        use_file_cache: bool = False,
        map_func: Optional[Callable[[datasets.Dataset], datasets.Dataset]] = None,
        rename_columns: Optional[Dict[str, str]] = None,
        cast_columns: Optional[Dict[str, Dict[str, Any]]] = None,
        data_format: Dict[str, Dict[str, Any]],
        seq_tag_column: Optional[str] = "id",
        sorting_seq_len_column_data: Optional[str] = None,
        sorting_seq_len_column: Optional[str] = None,
        **kwargs,
    ):
        """
        :param dataset_opts: either a dict of options for :func:`datasets.load_dataset`
            or a path to a local dataset for :func:`datasets.load_from_disk`,
            or a list of Arrow filenames to load with :func:`datasets.Dataset.from_file` and concatenate.
            It can also be a callable returning one of the above,
            or returning a :class:`datasets.Dataset` directly.
        :param use_file_cache: if True, will cache the dataset files on local disk using :mod:`file_cache`.
            This only works for dataset_opts which is a str or list of str (or callable returning that).
        :param map_func: optional function to apply to the dataset after loading
        :param rename_columns: if given, will rename these columns
        :param cast_columns: if given, will cast these columns to the specified types.
            This is useful if the dataset has not the expected types.
            See :func:`datasets.Dataset.cast` for details.
            You can also e.g. enforce some sample_rate for audio, etc.
        :param data_format:
            For each column name (data key), specify the format,
            as a dict with entries for "dim", "ndim", "shape", and/or "dtype",
            compatible to :class:`Tensor`.
            It can be a subset of the available columns.
            If "vocab" is specified, and the underlying HF datasets column is of dtype "string",
            it will automatically tokenize the string using the vocab.
        :param seq_tag_column: key (column name) in the dataset to use as sequence tag.
            If None, will use the sequence index as tag.
        :param sorting_seq_len_column_data: key (column name) in the dataset to use for sorting by sequence length.
            It will take len(dataset[sorting_seq_len_column_data]) as sequence length (only for sorting/shuffling).
        :param sorting_seq_len_column: key (column name) in the dataset to use for sorting by sequence length.
            It will take the value of dataset[sorting_seq_len_column] as sequence length (only for sorting/shuffling).
            E.g. some datasets provide "duration", "duration_ms", "wav_filesize" or similar such information
            which can be used.
        """
        super().__init__(**kwargs)

        self.dataset_opts = dataset_opts
        self.use_file_cache = use_file_cache
        self.map_func = map_func
        self.rename_columns = rename_columns
        self.cast_columns = cast_columns

        self.data_format: Dict[str, Tensor] = {k: _make_tensor_template(v, k) for k, v in data_format.items()}
        self.seq_tag_column: Optional[str] = seq_tag_column
        self.sorting_seq_len_column_data = sorting_seq_len_column_data
        self.sorting_seq_len_column = sorting_seq_len_column

        self.labels = {k: data.vocab.labels for k, data in self.data_format.items() if data.vocab}
        self.num_outputs = {k: (data.dim or 1, data.ndim) for k, data in self.data_format.items()}

        self.hf_dataset: Optional[datasets.Dataset] = None  # lazily loaded, _lazy_init
        self._seq_order: Optional[Sequence[int]] = None  # init_seq_order
        self._seq_tags: Optional[List[str]] = None  # get_all_tags cache

    def _lazy_init(self):
        if self.hf_dataset is not None:
            return

        # Load the dataset
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import datasets

        dataset_opts = self.dataset_opts
        if callable(dataset_opts):
            dataset_opts = dataset_opts()
        if self.use_file_cache:
            assert isinstance(dataset_opts, (str, os.PathLike, list, tuple)), (
                f"{self}: with use_file_cache, dataset_opts must be str or list of str, got {type(dataset_opts)}"
            )
            if isinstance(dataset_opts, (str, os.PathLike)):
                dataset_opts = get_arrow_shard_files_from_hf_dataset_dir(dataset_opts)
            assert isinstance(dataset_opts, (list, tuple))
            cache = file_cache.get_instance()
            dataset_opts = [cache.get_file(os.fspath(fn)) for fn in dataset_opts]
            self.set_file_cache(cache)
        if isinstance(dataset_opts, dict):
            self.hf_dataset = datasets.load_dataset(**dataset_opts)
        elif isinstance(dataset_opts, str):
            self.hf_dataset = datasets.load_from_disk(dataset_opts)
        elif isinstance(dataset_opts, (list, tuple)):
            self.hf_dataset = datasets.concatenate_datasets([datasets.Dataset.from_file(fn) for fn in dataset_opts])
        elif isinstance(dataset_opts, datasets.Dataset):
            self.hf_dataset = dataset_opts
        else:
            raise TypeError(f"{self}: invalid dataset_opts type {type(dataset_opts)}")
        assert isinstance(self.hf_dataset, datasets.Dataset), (
            f"{self}: Expected single dataset, got {type(self.hf_dataset)} {self.hf_dataset}. Specify split if needed."
        )

        if self.map_func is not None:
            self.hf_dataset = self.map_func(self.hf_dataset)

        if self.rename_columns:
            self.hf_dataset = self.hf_dataset.rename_columns(self.rename_columns)

        if self.cast_columns:
            # Note: prefer cast_column, as this can avoid using `map`, i.e. be faster.
            for key, column_format in self.cast_columns.items():
                assert key in self.hf_dataset.features, (
                    f"{self}: cast_column {key} not in dataset features {self.hf_dataset.features}"
                )
                feat = datasets.features.features.generate_from_dict(column_format)
                self.hf_dataset = self.hf_dataset.cast_column(key, feat)

        if self.seq_tag_column:
            assert self.seq_tag_column in self.hf_dataset.features, (
                f"{self}: seq_tag_column {self.seq_tag_column} not in dataset features {self.hf_dataset.features}"
            )
            assert self.hf_dataset.features[self.seq_tag_column].dtype in ("string", "int64"), (
                f"{self}: seq_tag_column {self.seq_tag_column} must be of dtype string or int64,"
                f" got {self.hf_dataset.features[self.seq_tag_column].dtype}"
            )

        selected_columns = list(self.data_format.keys())
        if self.seq_tag_column and self.seq_tag_column not in selected_columns:
            selected_columns.append(self.seq_tag_column)
        if self.sorting_seq_len_column and self.sorting_seq_len_column not in selected_columns:
            selected_columns.append(self.sorting_seq_len_column)
        if self.sorting_seq_len_column_data and self.sorting_seq_len_column_data not in selected_columns:
            selected_columns.append(self.sorting_seq_len_column_data)
        self.hf_dataset = self.hf_dataset.select_columns(selected_columns)

        self.hf_dataset.set_format("numpy")

        for key, user_format in self.data_format.items():
            feature = self.hf_dataset.features[key]
            inferred_format = _infer_data_format_for_feature(feature, f"{self}: column {key}: ")
            if user_format.vocab and inferred_format["dtype"] == "string":
                pass  # allow to auto-tokenize strings when vocab is specified
            else:
                for key_ in ["dtype", "ndim", "dim"]:
                    assert getattr(user_format, key_) == inferred_format[key_], (
                        f"{self}: column {key}, user-specified {user_format}, {key_}:"
                        f" user-specified {getattr(user_format, key_)} does not match inferred {inferred_format[key_]}"
                    )
            if "vocab" in inferred_format and not user_format.vocab:
                assert user_format.sparse, f"{self}: column {key}: user_format expected to be sparse, got {user_format}"
                user_format.sparse_dim.vocab = Vocabulary.create_vocab(**inferred_format["vocab"])
                self.labels[key] = user_format.vocab.labels

    def get_data_keys(self) -> List[str]:
        """:return: list of data keys"""
        return list(self.data_format.keys())

    def get_target_list(self) -> List[str]:
        """:return: list of target keys"""
        return self.get_data_keys()  # it's somewhat arbitrary...

    def get_data_shape(self, key: str) -> List[int]:
        """:return: data shape for the given key"""
        return list(self.data_format[key].shape)

    def get_data_dim(self, key: str) -> int:
        """:return: data dimension for the given key"""
        return self.data_format[key].dim

    def is_data_sparse(self, key: str) -> bool:
        """:return: whether the data is sparse for the given key"""
        return self.data_format[key].sparse

    def get_data_dtype(self, key: str) -> str:
        """:return: dtype"""
        return self.data_format[key].dtype

    def _get_seq_len(self, seq_idx: int) -> Union[int, float]:
        if self._seq_order_seq_lens_by_idx is not None:
            self._get_seq_len = self._seq_order_seq_lens_by_idx.__getitem__  # faster
            return self._seq_order_seq_lens_by_idx[seq_idx]
        assert not self._seq_order_seq_lens_file  # not expected to call this
        if self.sorting_seq_len_column:
            self._seq_order_seq_lens_by_idx = numpy.array(self.hf_dataset[self.sorting_seq_len_column])
            self._get_seq_len = self._seq_order_seq_lens_by_idx.__getitem__  # faster
            v = self._seq_order_seq_lens_by_idx[seq_idx]
            return int(v)  # noqa
        if self.sorting_seq_len_column_data:
            v = self.hf_dataset[seq_idx][self.sorting_seq_len_column_data]
            return len(v)  # noqa
        raise ValueError(
            f"{self}: sorting/shuffling by seq len not configured,"
            f" need sorting_seq_len_column or sorting_seq_len_column_data"
        )

    @property
    def num_seqs(self) -> int:
        """:return: number of sequences"""
        assert self._seq_order is not None, "num_seqs is only known after calling init_seq_order()"
        return len(self._seq_order)

    def get_tag(self, sorted_seq_idx: int) -> str:
        """:return: tag of the sequence"""
        corpus_seq_idx = self.get_corpus_seq_idx(sorted_seq_idx)
        self._lazy_init()
        dataset_item = self.hf_dataset[corpus_seq_idx]
        return self._get_seq_tag(corpus_seq_idx, dataset_item)

    def get_all_tags(self) -> List[str]:
        """:return: all tags"""
        if self._seq_tags is not None:
            return self._seq_tags
        self._lazy_init()
        if self.seq_tag_column:
            res = list(map(str, self.hf_dataset[self.seq_tag_column]))
        else:
            res = [f"seq-{i}" for i in range(self.hf_dataset.num_rows)]
        self._seq_tags = res
        return res

    def get_total_num_seqs(self, *, fast: bool = False) -> int:
        """:return: total number of sequences in the dataset"""
        if fast:
            return super().get_total_num_seqs(fast=True)
        self._lazy_init()
        return self.hf_dataset.num_rows

    def init_seq_order(
        self,
        epoch: Optional[int] = None,
        seq_list: Optional[Sequence[str]] = None,
        seq_order: Optional[Sequence[int]] = None,
    ) -> bool:
        """
        :param epoch:
        :param seq_list: List of sequence tags, to set a predefined order.
        :param seq_order: List of corpus sequence indices, to set a predefined order.
        :returns whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if seq_order is not None:
            self._seq_order = seq_order
        elif seq_list is not None:
            all_tags = self.get_all_tags()
            self._seq_order = [all_tags.index(tag) for tag in seq_list]
        elif epoch is None:
            self._seq_order = ()
        else:
            self._lazy_init()
            self._seq_order = self.get_seq_order_for_epoch(
                epoch=epoch, num_seqs=self.hf_dataset.num_rows, get_seq_len=self._get_seq_len
            )
        return True

    def _collect_single_seq(self, seq_idx: int) -> DatasetSeq:
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import datasets

        corpus_seq_idx = self.get_corpus_seq_idx(seq_idx)

        def _ensure_numpy(k, x):
            if isinstance(x, numpy.ndarray):  # fast path
                return x
            if isinstance(x, str):
                if self.data_format[k].dtype == "string":
                    return str_to_numpy_array(x)
                if self.data_format[k].vocab:
                    return numpy.array(self.data_format[k].vocab.get_seq(x), dtype=self.data_format[k].dtype)
                raise ValueError(f"{self}: column {k}: cannot convert string {x!r} to numpy array")
            feat = self.hf_dataset.features[k]
            if isinstance(feat, datasets.features.Audio):
                # In HF datasets 3, this is just a dict.
                # In HF datasets 4, this can also be a datasets.features._torchcodec.AudioDecoder.
                assert isinstance(x, dict) or x.__class__.__name__ == "AudioDecoder"
                if feat.decode:
                    x = x["array"]
                else:
                    x = x["bytes"]
            if isinstance(x, numpy.ndarray):  # fast path
                return x
            if isinstance(x, (bytes, bytearray)):
                return numpy.frombuffer(x, dtype=self.data_format[k].dtype)
            return numpy.array(x)

        self._lazy_init()
        dataset_item = self.hf_dataset[corpus_seq_idx]
        seq_tag = self._get_seq_tag(corpus_seq_idx, dataset_item)
        features = {k: _ensure_numpy(k, dataset_item[k]) for k in self.data_format}
        return DatasetSeq(seq_idx, features=features, seq_tag=seq_tag)

    def _get_seq_tag(self, corpus_seq_idx: int, dataset_item: Dict[str, Any]) -> str:
        if self.seq_tag_column:
            seq_tag = dataset_item[self.seq_tag_column]
            assert isinstance(seq_tag, (str, int, numpy.int64)), f"got {type(seq_tag)} {seq_tag!r}"
            seq_tag = str(seq_tag)
        else:
            seq_tag = f"seq-{corpus_seq_idx}"
        return seq_tag

    def get_current_seq_order(self) -> Sequence[int]:
        """:return: list of corpus seq idx"""
        assert self._seq_order is not None
        return self._seq_order

    def get_corpus_seq_idx(self, sorted_seq_idx: int) -> int:
        """:return: corpus seq idx"""
        return int(self._seq_order[sorted_seq_idx])


def get_arrow_shard_files_from_hf_dataset_dir(hf_data_dir: Union[str, os.PathLike]) -> List[str]:
    """
    Given some HF datasets directory (created via :func:`datasets.save_to_disk`),
    return the list of Arrow shard files (``data-*-of-*.arrow``).
    This also verifies that the directory looks like a valid HF datasets directory.
    The order of the returned list is by shard index.
    Note that this does not load the dataset, just inspects the directory structure.

    :param hf_data_dir: directory
    :return: list of Arrow shard files
    """
    hf_data_dir = os.fspath(hf_data_dir)
    content = os.listdir(hf_data_dir)
    assert "state.json" in content, f"not a valid HF datasets dir: {hf_data_dir!r}"
    assert "dataset_info.json" in content, f"not a valid HF datasets dir: {hf_data_dir!r}"
    pat = re.compile("^(.*)-([0-9]+)-of-([0-9]+).arrow$")
    content = [pat.match(fn) for fn in content]
    content = [m for m in content if m]
    assert content, f"no matching .arrow files in {hf_data_dir!r} found, expected *-*-of-*.arrow"
    prefix = content[0].group(1)
    assert all(m.group(1) == prefix for m in content), (
        f"mismatching prefix in {hf_data_dir!r}, expected {prefix}, got {[m.group(1) for m in content]}"
    )
    num_shards = int(content[0].group(3))
    assert all(int(m.group(3)) == num_shards for m in content), (
        f"mismatching number of shards in {hf_data_dir!r}, expected {num_shards}, got {[m.group(3) for m in content]}"
    )
    assert len(content) == num_shards, f"expected {num_shards} shard files in {hf_data_dir!r}, got {content}"
    content_by_idx = {int(m.group(2)): m for m in content}
    assert set(content_by_idx.keys()) == set(range(num_shards)), (
        f"expected shard indices 0..{num_shards - 1} in {hf_data_dir!r}, got {sorted(content_by_idx.keys())}"
    )
    return [hf_data_dir + "/" + content_by_idx[i].group(0) for i in range(num_shards)]


def _infer_data_format_for_feature(
    feature: Union[
        datasets.features.Sequence,
        datasets.features.ClassLabel,
        datasets.features.Value,
        datasets.features.Array2D,
        datasets.features.Array3D,
        datasets.features.Array4D,
        datasets.features.Audio,
    ],
    exc_prefix: str = "",
) -> Dict[str, Any]:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import datasets

    labels = None
    num_classes = None
    num_dims = 0
    while isinstance(feature, datasets.features.Sequence):
        feature: datasets.features.List  # typing for HF datasets 4
        num_dims += 1
        if feature.length != -1:
            num_classes = feature.length
        feature = feature.feature
    if isinstance(feature, datasets.features.ClassLabel):
        labels = feature.names
        dtype = feature.dtype
        num_classes = feature.num_classes  # noqa
    elif isinstance(feature, datasets.features.Value):
        dtype = feature.dtype
    elif isinstance(feature, (datasets.features.Array2D, datasets.features.Array3D, datasets.features.Array4D)):
        dtype = feature.dtype
        num_classes = feature.shape[-1]
        num_dims += len(feature.shape)
    elif isinstance(feature, datasets.features.Audio):
        if feature.decode:
            dtype = "float32"  # samples
        else:
            dtype = "uint8"  # bytes
        num_dims += 1  # time axis
    else:
        assert False, f"{exc_prefix}unsupported feature type {type(feature)} {feature}"

    d = {"dim": num_classes, "ndim": num_dims, "dtype": dtype}
    if labels:
        d["sparse"] = True
        d["vocab"] = {"vocab_file": None, "labels": labels, "unknown_label": None}
    return d


def _make_tensor_template(data: Union[Dict[str, Any], Tensor], name: str) -> Tensor:
    if isinstance(data, Tensor):
        data = data.copy(name)
    else:
        assert isinstance(data, dict)
        data = Tensor(name, batch_dim_axis=None, **data)
    assert data.batch_dim_axis is None
    return data



#!/usr/bin/env python3

"""
Dump translation dataset to HDF.
"""

from __future__ import annotations

import sys
import argparse
import gzip
import pickle
import typing

import numpy
import h5py
from operator import itemgetter
from itertools import islice, zip_longest

UNKNOWN_LABEL = "<UNK>"
POSTFIX = "</S>"


class HDFTranslationDatasetCreator(object):
    """
    Creates the same HDF file as hdf_dump.py called on a TranslationDataset, but does so much faster
    and using much less memory.

    Input parameters are basically the files in a TranslationDataset folder. Additionally, the number of lines has to
    be given.
    """

    def __init__(
        self,
        hdf_file,
        source_file,
        target_file,
        source_vocabularies,
        target_vocabularies,
        source_factors,
        target_factors,
        number_of_lines,
        factor_separator="|",
        compression_method=None,
        line_buffer_size=100000,
        data_buffer_size=5000000,
    ):
        """
        :param str hdf_file: filename for the hdf file being created
        :param str source_file: filename of the source text file
        :param str target_file: filename of the target text file
        :param list[str] source_vocabularies: Filenames of the source vocabularies (in pickle format).
            Usually there is just one. In case of source factors, provide one per factor.
        :param list[str] target_vocabularies: Filenames of the target vocabularies (in pickle format).
            Usually there is just one. In case of target factors, provide one per factor.
        :param list[str] source_factors: Data keys for the source factors. First data key is always 'data'
            and must not be contained in the list.
        :param list[str] target_factors: Data keys for the target factors. First data key is always 'classes'
            and must not be contained in the list.
        :param str factor_separator: "String used to separate factors of the words.
            E.g. if "|", words are expected to be of format "<lemma>|<factor>|...".
        :param int number_of_lines: line count in source_file and target_file
        :param str compression_method: Optional compression method as supported by h5py.File.create_dataset().
            Applied to the main data only ('inputs' and 'target/data/classes').
        :param int line_buffer_size: number of corpus lines to read and process at once
        :param int data_buffer_size: space to reserve inside the hdf file at once, in numbers of integers
        """

        self.hdf_file_name = hdf_file
        self.hdf_file = None
        self.compression_method = compression_method

        self.source_file_handle = self._open_file(source_file)
        self.target_file_handle = self._open_file(target_file)

        source_vocabularies = [self._read_vocabulary(v) for v in source_vocabularies]
        target_vocabularies = [self._read_vocabulary(v) for v in target_vocabularies]

        self.source_data_keys = ["data"] + source_factors
        self.target_data_keys = ["classes"] + target_factors

        self._vocabularies = {"source": source_vocabularies, "target": target_vocabularies}
        self._vocabulary_sizes = {
            "source": [len(v) for v in source_vocabularies],
            "target": [len(v) for v in target_vocabularies],
        }
        self._unknown_ids = {
            "source": [v.get(UNKNOWN_LABEL) for v in source_vocabularies],
            "target": [v.get(UNKNOWN_LABEL) for v in target_vocabularies],
        }

        self.number_of_lines = number_of_lines
        self.line_buffer_size = line_buffer_size
        self.data_buffer_size = data_buffer_size

        self._number_of_processed_lines = 0
        self._write_offsets = {data_key: 0 for data_key in self.source_data_keys + self.target_data_keys}

        self.factor_separator = factor_separator

    def create(self):
        """
        Main function writing the HDF file.
        """
        self._init_hdf_file()

        print("Setting attributes...", file=sys.stderr)
        sys.stderr.flush()
        self._write_attributes()
        print("Done.", file=sys.stderr)

        print("Writing labels (vocabulary)...", file=sys.stderr)
        sys.stderr.flush()
        self._write_labels()
        print("Done.", file=sys.stderr)

        print("Writing source, target, sequence lengths and tags for all lines...", file=sys.stderr)
        sys.stderr.flush()
        end_of_file = False
        while not end_of_file:
            end_of_file = self._write_data()
            if not end_of_file:
                print("> Processed {} lines.".format(self._number_of_processed_lines), file=sys.stderr)
                sys.stderr.flush()
        print("Done.", file=sys.stderr)
        sys.stderr.flush()

        self.hdf_file.close()

    def _init_hdf_file(self):
        """
        Sets up the HDF file and initializes the datasets that will be filled.
        """
        self.hdf_file = h5py.File(self.hdf_file_name, "w")

        self.hdf_file.create_group("targets/data")
        self.hdf_file.create_group("targets/size")
        self.hdf_file.create_group("targets/labels")

        num_data_keys = len(self.source_data_keys) + len(self.target_data_keys)
        self.hdf_file.create_dataset("seqLengths", (self.number_of_lines, num_data_keys), dtype="int32")
        max_tag_length = len("line-") + len(str(self.number_of_lines))
        self._tag_dtype = "S{}".format(max_tag_length)
        self.hdf_file.create_dataset("seqTags", (self.number_of_lines,), dtype=self._tag_dtype)

        self.hdf_file.create_dataset(
            "inputs", (self.data_buffer_size,), maxshape=(None,), dtype="int32", compression=self.compression_method
        )

        # HDF format expects one input only, so store source factors as target too.
        for data_key in self.source_data_keys[1:] + self.target_data_keys:
            self.hdf_file["targets/data"].create_dataset(
                data_key, (self.data_buffer_size,), maxshape=(None,), dtype="int32", compression=self.compression_method
            )

    def _write_attributes(self):
        """
        Writes several attributes to the HDF file.
        """
        for index, data_key in enumerate(self.source_data_keys[1:], start=1):
            self.hdf_file["targets/size"].attrs[data_key] = (self._vocabulary_sizes["source"][index], 1)
        for index, data_key in enumerate(self.target_data_keys):
            self.hdf_file["targets/size"].attrs[data_key] = (self._vocabulary_sizes["target"][index], 1)

        # Those should be deprecated, but include nevertheless to exactly reproduce hdf_dump.
        self.hdf_file.attrs["inputPattSize"] = self._vocabulary_sizes["source"][0]
        self.hdf_file.attrs["numLabels"] = self._vocabulary_sizes["target"][0]

    def _write_labels(self):
        """
        Writes the labels (i.e. target vocabulary) to the HDF file.
        """
        for side in ["source", "target"]:
            # We have to write it for the source factors too, because they are treated like targets.

            data_keys = self.source_data_keys if side == "source" else self.target_data_keys
            for index, data_key in enumerate(data_keys):
                if data_key == "data":
                    continue

                sorted_vocabulary_tuples = sorted(self._vocabularies[side][index].items(), key=itemgetter(1))
                labels = [word.encode("utf8") for (word, _) in sorted_vocabulary_tuples]

                assert len(labels) == self._vocabulary_sizes[side][index], "Word ids were not unique."

                max_label_length = max([len(label) for label in labels])
                dtype = "S{}".format(max_label_length + 1)

                labels = [numpy.array(label, dtype=dtype, ndmin=1) for label in labels]
                labels = numpy.concatenate(labels)

                self.hdf_file["targets/labels"].create_dataset(
                    data_key, (self._vocabulary_sizes[side][index],), data=labels, dtype=dtype
                )

    def _write_data(self):
        """
        Loads a chunk of lines from the corpus and writes all corresponding data to the HDF file.

        :return: whether the end of the corpus is reached
        :rtype: bool
        """
        data_chunks = self._get_chunk()

        if not data_chunks["data"]:
            self._finalize_data()
            return True

        source_lengths = [len(line) for line in data_chunks["data"]]
        target_lengths = [len(line) for line in data_chunks["classes"]]

        chunk_size = len(data_chunks["data"])

        self._write_lengths(source_lengths, target_lengths)
        self._write_tags(chunk_size)

        for data_key, data_chunk in data_chunks.items():
            self._write_data_indices(data_chunk, data_key)

        self._number_of_processed_lines += chunk_size

        return False

    def _get_chunk(self):
        """
        Reads in the next chunk of lines from the corpus files.

        :return: a dict 'data_key' -> word indices for 'data_key' (int32, shape [Lines]) for all source and target data
        :rtype: dict[str,list[numpy.ndarray]]
        """
        source_lines = islice(self.source_file_handle, self.line_buffer_size)
        target_lines = islice(self.target_file_handle, self.line_buffer_size)

        data_chunks = {data_key: [] for data_key in self.source_data_keys + self.target_data_keys}
        for source_line, target_line in zip_longest(source_lines, target_lines):
            assert source_line is not None and target_line is not None, "Number of source and target lines differ."

            source_indices = self._line_to_indices(source_line, "source")
            for data_key, indices in zip(self.source_data_keys, source_indices):
                data_chunks[data_key].append(indices)

            target_indices = self._line_to_indices(target_line, "target")
            for data_key, indices in zip(self.target_data_keys, target_indices):
                data_chunks[data_key].append(indices)

        return data_chunks

    def _write_lengths(self, source_lengths, target_lengths):
        """
        Writes the sequence lengths to the HDF file.

        :param list[int] source_lengths: lengths of all source lines in current chunk
        :param list[int] target_lengths: lengths of all target lines in current chunk
        """
        # We treat source factors as targets internally because HDF format does not support multiple inputs.

        # For each sequence, seqLengths is expected to contain the length of the input and the length of each of the
        # targets, in alphabetical order of the target data keys. As all source and all target factors share the
        # same lengths we just choose between the source and target lengths.
        target_sequence_lengths = {}
        for data_key in self.source_data_keys[1:]:
            target_sequence_lengths[data_key] = source_lengths
        for data_key in self.target_data_keys:
            target_sequence_lengths[data_key] = target_lengths

        # Now sort by key.
        key_lengths_tuples_sorted = sorted(target_sequence_lengths.items(), key=lambda x: x[0])  # type: typing.List[typing.Tuple[str,typing.List[int]]]  # nopep8
        target_lengths_sorted = [key_length_tuple[1] for key_length_tuple in key_lengths_tuples_sorted]

        # Finally, add one time the source lengths for the input ("data") and convert to numpy.
        lengths = numpy.array([source_lengths] + target_lengths_sorted, dtype="int32").transpose()

        offset = self._number_of_processed_lines
        assert len(lengths) + offset <= self.number_of_lines, "More lines in the corpus files than specified."

        self.hdf_file["seqLengths"][offset : offset + len(lengths), :] = lengths

    def _write_tags(self, chunk_size):
        """
        Writes the sequence tags to the HDF file.

        :param int chunk_size: number of lines in the current chunk
        """
        offset = self._number_of_processed_lines

        tags = [numpy.array("line-" + str(offset + i), dtype=self._tag_dtype, ndmin=1) for i in range(chunk_size)]
        tags = numpy.concatenate(tags)

        self.hdf_file["seqTags"][offset : offset + chunk_size] = tags

    def _write_data_indices(self, chunk, data_key):
        """
        Writes the main data (word indices for the source or target corpus) to the HDF file.

        :param list[numpy.ndarray] chunk: word indices for all lines in the current chunk
        :param str data_key: "data", "classes" or a name of a factor
        """
        indices = numpy.concatenate(chunk)

        if data_key == "data":
            dataset = self.hdf_file["inputs/"]
        else:
            dataset = self.hdf_file["targets/data/{}".format(data_key)]

        offset = self._write_offsets[data_key]
        length = len(indices)
        if offset + length > len(dataset):
            buffer_size = max(self.data_buffer_size, length)
            dataset.resize((offset + buffer_size,))

        dataset[offset : offset + length] = indices

        self._write_offsets[data_key] += length

    def _finalize_data(self):
        """
        Called after all data is written. Checks number of lines and resizes datasets down to actual data size.
        """
        # Make sure the number of lines given by the user was correct.
        # Otherwise lengths and labels would have trailing zeros.
        assert self.number_of_lines == self._number_of_processed_lines, (
            "Fewer lines ({}) in the corpus files than specified ({}).".format(
                self._number_of_processed_lines, self.number_of_lines
            )
        )

        # Trim datasets to actually occupied length, i.e. remove unused reserved space.
        self.hdf_file["inputs"].resize((self._write_offsets["data"],))
        for data_key in self.source_data_keys[1:] + self.target_data_keys:
            self.hdf_file["targets/data/{}".format(data_key)].resize((self._write_offsets[data_key],))

    def _line_to_indices(self, line, side):
        """
        Converts a line of text to arrays of word indices.

        :param str line: input line
        :param str side: "source" or "target"
        :return: word indices (int32, shape [num_words]) for all source or target factors
        :rtype: list[numpy.ndarray]
        """
        data_keys = self.source_data_keys if side == "source" else self.target_data_keys

        words = line.strip().split()

        if len(data_keys) == 1:
            word_list_per_factor = [words + [POSTFIX]]
        else:
            if words:
                words_split_into_factors = [word.split(self.factor_separator) for word in words]
                assert all(len(factors) == len(data_keys) for factors in words_split_into_factors), (
                    "All words must have all factors. Expected: " + self.factor_separator.join(data_keys)
                )
                word_list_per_factor = zip(*words_split_into_factors)
            else:
                word_list_per_factor = [[]] * len(data_keys)
            word_list_per_factor = [list(words) + [POSTFIX] for words in word_list_per_factor]

        indices_list = []
        for index, (data_key, words) in enumerate(zip(data_keys, word_list_per_factor)):
            indices = [self._vocabularies[side][index].get(word, self._unknown_ids[side][index]) for word in words]

            indices_numpy = numpy.array(indices, dtype=numpy.int32)
            indices_list.append(indices_numpy)

        return indices_list

    @staticmethod
    def _open_file(file_name):
        """
        :param str file_name: filename of a plain text file, possibly zipped
        :return: file handle
        :rtype: io.TextIOWrapper|gzip.GzipFile
        """
        if file_name.endswith(".gz"):
            return gzip.open(file_name, "rt")
        else:
            return open(file_name, "r")

    @staticmethod
    def _read_vocabulary(file_name):
        """
        :param str file_name: filename of the vocabulary (in pickle format)
        :return: mapping from words to indices
        :rtype: dict[str,int]
        """
        file_handle = open(file_name, "rb")
        vocabulary = pickle.load(file_handle)
        return vocabulary


def parse_args():
    """
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--source_corpus", required=True, help="Source corpus file, possibly zipped.")
    parser.add_argument("-t", "--target_corpus", required=True, help="Target corpus file, possibly zipped.")
    parser.add_argument(
        "-v",
        "--source_vocabulary",
        required=True,
        help="Source vocabulary in pickle format."
        "In case of source factors provide a comma separated list containing vocabularies for each factor.",
    )
    parser.add_argument(
        "-w",
        "--target_vocabulary",
        required=True,
        help="Target vocabulary in pickle format."
        "In case of target factors provide a comma separated list containing vocabularies for each factor.",
    )
    parser.add_argument("-o", "--hdf_file", required=True, help="Output HDF file name.")
    parser.add_argument("--source_factors", help="Comma separated list of data keys for the source factors.")
    parser.add_argument("--target_factors", help="Comma separated list of data keys for the target factors.")
    parser.add_argument(
        "--factor_separator",
        default="|",
        help="String used to separate factors of the words, E.g. if '|', words are expected to be "
        "of format '<lemma>|<factor>|...'",
    )
    parser.add_argument(
        "-n", "--number_of_lines", required=True, type=int, help="The number of total lines in the corpus files."
    )
    parser.add_argument(
        "-c", "--compression", help="Type of compression (e.g. 'gzip', 'lzf'). Turned off if not given."
    )
    parser.add_argument("-l", "--line_buffer_size", type=int, help="How many lines to read at once.", default=100000)
    parser.add_argument(
        "-d",
        "--data_buffer_size",
        type=int,
        help="How much space to reserve in the HDF dataset at once (in number of integers).",
        default=5000000,
    )

    return parser.parse_args()


def main():
    """
    Main entry.
    """
    args = parse_args()

    # In case of source or target factors we need a vocabularies for each.
    source_vocabularies = args.source_vocabulary.split(",")
    target_vocabularies = args.target_vocabulary.split(",")

    source_factors = args.source_factors.split(",") if args.source_factors else []
    target_factors = args.target_factors.split(",") if args.target_factors else []

    assert len(source_factors) + 1 == len(source_vocabularies), (
        "Number of source factors must be one less "
        "than number of source vocabularies (first factor is always called 'data')"
    )
    assert len(target_factors) + 1 == len(target_vocabularies), (
        "Number of target factors must be one less "
        "than number of target vocabularies (first factor is always called 'classes')"
    )

    HDFTranslationDatasetCreator(
        args.hdf_file,
        args.source_corpus,
        args.target_corpus,
        source_vocabularies,
        target_vocabularies,
        source_factors,
        target_factors,
        args.number_of_lines,
        args.factor_separator,
        args.compression,
        args.line_buffer_size,
        args.data_buffer_size,
    ).create()


if __name__ == "__main__":
    sys.exit(main())



"""
Utilities for dimension tags, dimensions, axes.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, TypeVar, Sequence, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import get_backend_by_tensor, global_backend

T = TypeVar("T")

__all__ = [
    "range_over_dim",
    "range_over_dim_strided",
    "range_over_merged_dims",
    "linspace_over_dim",
    "replace_dim",
    "replace_dim_v2",
    "set_sparse_dim",
    "dim_match_priority_when_needed",
    "num_elements_of_shape",
    "masked_fraction_of_shape",
    "last_frame_position_of_dim",
    "use_mask_default",
]

if TYPE_CHECKING:
    from returnn.config import Config


def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor[T]:
    """
    :param dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim]
    """
    if dim.dyn_size_ext is not None:
        backend = get_backend_by_tensor(dim.dyn_size_ext, fallback=global_backend)
    else:
        backend = global_backend
    return backend.range_over_dim(dim, dtype=dtype, device=device)


def range_over_dim_strided(
    dim: Dim,
    *,
    stride: Union[int, Tensor],
    out_dim: Optional[Dim] = None,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> Tuple[Tensor[T], Dim]:
    """
    :param dim:
    :param stride:
    :param out_dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim], out_dim
    """
    if out_dim is None:
        out_dim = dim.ceildiv_right(stride)
    return rf.range_over_dim(out_dim, dtype=dtype, device=device) * stride, out_dim


def range_over_merged_dims(
    dims: Sequence[Dim], *, dtype: Optional[str] = None, device: Optional[str] = None
) -> Tensor[T]:
    """
    This is if you want to index into a merged dim.
    Related: :func:`rf.merge_dims`.

    :param dims:
    :param dtype:
    :param device:
    :return: tensor with shape [dim_0, ..., dim_n] -> sparse_dim = merged_dim, where merged_dim = dim_0 * ... * dim_n
    """
    assert len(dims) >= 1
    merged_dim = dims[0]
    for dim in dims[1:]:
        merged_dim *= dim
    indices = rf.range_over_dim(merged_dim, dtype=dtype, device=device)
    if len(dims) > 1:
        indices = rf.split_dims(indices, axis=merged_dim, dims=dims)
    return indices


def linspace_over_dim(
    dim: Dim,
    start: Union[float, Tensor] = 0.0,
    end: Union[float, Tensor] = 1.0,
    *,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> Tensor:
    """
    Linearly spaced values over a dim.

    :param dim: dim to range over
    :param start: start value
    :param end: end value
    :param dtype: dtype of the output tensor
    :param device: device of the output tensor
    :return: tensor with shape [dim] containing linearly spaced values between start and end
    """
    if dtype is None:
        dtype = rf.get_default_float_dtype()
    indices = rf.range_over_dim(dim, dtype=dtype, device=device)
    linspace = indices / rf.cast(rf.maximum(dim.get_size_tensor(device=indices.device), 1), dtype=indices.dtype)
    space_len = end - start
    if not isinstance(space_len, (int, float)) or space_len != 1:
        linspace *= space_len
    if not isinstance(start, (int, float)) or start != 0:
        linspace += start
    return linspace


def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Optional[Dim] = None) -> Tuple[Tensor, Dim]:
    """
    Also see: :func:`replace_dim_v2`, :func:`rf.merge_dims`, :func:`rf.split_dims`.

    :param source:
    :param in_dim: should be in ``source.dims``, to be replaced.
        If you want to replace the ``source.sparse_dim``, see :func:`set_sparse_dim`.
    :param out_dim: If not given, will create a new dim with the same size as ``in_dim``.
        Note: If the size of ``out_dim`` is different from ``in_dim``,
        currently the dim tag is replaced and there is no error -- this is not checked.
    :return: ``source`` with ``in_dim`` replaced by ``out_dim``, and ``out_dim``.
    """
    if not out_dim:
        out_dim = in_dim.copy(same_as_self=False, description="new-dim")
    # noinspection PyProtectedMember
    return source._raw_backend.replace_dim(source, in_dim=in_dim, out_dim=out_dim), out_dim


def replace_dim_v2(
    source: Tensor, *, in_dim: Dim, out_dim: Dim, allow_expand: bool = True, allow_shrink: bool = True
) -> Tensor:
    """
    Extends :func:`replace_dim` by also allowing to expand or shrink the dim
    (or rather, to not ignore this; when :func:`replace_dim` is used on a dim with different size,
     it will ignore this and anyway accept the new dim tag (currently)).

    :param source:
    :param in_dim: should be in ``source.dims``, to be replaced.
        If you want to replace the ``source.sparse_dim``, see :func:`set_sparse_dim`.
    :param out_dim: should not be in ``source.dims``, to be replaced.
        Note: In contrast to :func:`replace_dim`, you must provide this explicitly.
    :param allow_expand: if True, allow to expand the dim, i.e. if ``out_dim.size > in_dim.size``.
    :param allow_shrink: if True, allow to shrink the dim, i.e. if ``out_dim.size < in_dim.size``.
    :return: ``source`` with ``in_dim`` replaced by ``out_dim``.
    """
    if not rf.is_executing_eagerly():
        raise NotImplementedError  # just not implemented yet. we can do via :func:`cond`
    if in_dim not in source.dims:
        raise ValueError(f"replace_dim_v2: dim {in_dim} not in {source}")
    if out_dim in source.dims:
        raise ValueError(f"replace_dim_v2: dim {out_dim} already in {source}")
    old_size = in_dim.get_dim_value()
    new_size = out_dim.get_dim_value()
    if new_size == old_size:
        res, _ = rf.replace_dim(source, in_dim=in_dim, out_dim=out_dim)
    elif new_size > old_size:
        if not allow_expand:
            raise ValueError(
                f"replace_dim_v2: not allowed to expand: {old_size} -> {new_size},"
                f" for {in_dim=} {out_dim=}, in {source=}"
            )
        res, _ = rf.pad(
            source,
            axes=[in_dim],
            padding=[(0, out_dim.get_dim_value_tensor() - in_dim.get_dim_value_tensor())],
            out_dims=[out_dim],
            value=0,
        )
    else:
        if not allow_shrink:
            raise ValueError(
                f"replace_dim_v2: not allowed to shrink: {old_size} -> {new_size},"
                f" for {in_dim=} {out_dim=}, in {source=}"
            )
        res, _ = rf.slice(source, axis=in_dim, size=out_dim)
    return res


def set_sparse_dim(source: Tensor, sparse_dim: Dim) -> Tensor:
    """
    :param source:
    :param sparse_dim:
    :return: source with sparse_dim set
    """
    # noinspection PyProtectedMember
    return source._raw_backend.set_sparse_dim(source, sparse_dim)


def dim_match_priority_when_needed(dim: Dim, *other_dims: Dim) -> Dim:
    """
    :return: maybe copy of dim with higher match_priority if needed to distinguish from other_dims

    Why or when is this needed?

    For activation values, this should never be needed,
    and all dims should be unique.

    In case of self-attention, the standard way is to create a separate distinct dim
    to perform the attention reduction over.
    See :class:`SelfAttention`.

    However, in case of weight matrices, it is not unusual to have the same dim for both the input and output,
    so a square weight matrix.
    When reduction is performed in :func:`matmul`, we want to match the input feature dim
    to the dim in the weight matrix with higher match priority.

    So :func:`dim_match_priority_when_needed` would be applied on the input feature dim.

    https://github.com/rwth-i6/returnn/pull/871
    https://github.com/rwth-i6/returnn_common/issues/17#issuecomment-997463222
    """
    if dim in other_dims:
        return dim.copy(match_priority=1)
    return dim


def num_elements_of_shape(
    dims: Union[Dim, Sequence[Dim]], *, use_mask: bool = True, device: Optional[str] = None
) -> Union[int, Tensor]:
    """
    :param dims:
    :param use_mask:
    :param device: only for the case when we return a Tensor. by default, this is CPU (just as the size tensor).
    :return: num elements of a tensor of shape dims, properly considering masking
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not use_mask:
        n = 1
        for dim in dims:
            n *= dim.get_dim_value_tensor()
        return n
    if all(dim.is_static() for dim in dims):
        n = 1
        for dim in dims:
            n *= dim.size
        return n
    if len(dims) == 1:
        return dims[0].get_size_tensor(device=device)

    n: Union[int, Tensor] = 1
    postponed_dims = []
    for i, dim in enumerate(dims):
        # E.g. if dim==B, and some other dim dyn_size_ext has B, then we need to postpone this.
        related_dims = []
        for j, dim_ in enumerate(dims):
            if i == j:
                continue
            if dim_.dyn_size_ext is not None and dim in dim_.dyn_size_ext.dims:
                related_dims.append(dim_)
        if not related_dims:
            if dim.is_static():
                n *= dim.size
            else:
                n *= dim.get_size_tensor(device=device)
        else:
            postponed_dims.append(dim)
    if postponed_dims:
        n: Tensor
        n = rf.reduce_sum(n, axis=postponed_dims)
    return n


def masked_fraction_of_shape(dims: Union[Dim, Sequence[Dim]], *, inverse: bool = False) -> Union[int, float, Tensor]:
    """
    :param dims:
    :param inverse: if True, return the inverse of the fraction
    :return: :func:`num_elements_of_shape`(dims) / prod(dims) if not inverse else prod(dims) / num_elements
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not any(dim.need_masking() for dim in dims):
        return 1
    num_elems_masked = num_elements_of_shape(dims)
    num_elems_total = 1
    for dim in dims:
        num_elems_total *= dim.get_dim_value_tensor()
    return (num_elems_masked / num_elems_total) if not inverse else (num_elems_total / num_elems_masked)


def last_frame_position_of_dim(
    dim: Dim, *, device: Optional[str] = None, allow_scalar_on_cpu: bool = True
) -> Union[int, Tensor]:
    """
    :param dim:
    :param device:
    :param allow_scalar_on_cpu: if device is not given, we would use rf.get_default_device() for dynamic sizes.
        If this is True, we would allow to return a scalar dynamic size on CPU.
    :return: last frame position of dim
    """
    if dim.size is not None:
        if dim.size > 0:
            return dim.size - 1
        return 0
    if device is None:
        if allow_scalar_on_cpu and not dim.dyn_size_ext.dims:
            device = "cpu"
        else:
            device = rf.get_default_device()
    pos = dim.get_dyn_size_ext_for_device(device) - 1
    pos = rf.maximum(pos, 0)
    pos.sparse_dim = dim
    return pos


def use_mask_default(
    *,
    default: Optional[bool] = None,
    default_false_for_behavior_version_up_to: Optional[int] = None,
    func_name: Optional[str] = None,
) -> Optional[bool]:
    """
    Check the global RETURNN config for the ``rf_use_mask``
    on what default we should use for the ``use_mask`` argument in various functions
    (e.g. :func:`conv`, :func:`pool`, :func:`reduce`, :func:`matmul`, ...).

    See issue `#1691 <https://github.com/rwth-i6/returnn/issues/1691>`__.

    :param default: what to return if it is not defined in the config,
        and ``default_false_for_behavior_version_up_to`` does not apply.
    :param default_false_for_behavior_version_up_to: if it is not defined in the config,
        and if this is set, and the behavior version is less or equal,
        then return False by default, i.e. do not use the mask by default, if it is not defined in the config.
        This takes precedence over `default`.
    :param func_name: if specified, also check
    :return: what to use for the ``use_mask`` argument by default
    """
    from returnn.config import get_global_config

    config = get_global_config(raise_exception=False)
    if config:
        config_value = _get_opt_bool_from_config(config, "rf_use_mask")
        if config_value is not None:
            return config_value
        if func_name:
            config_value = _get_opt_bool_from_config(config, f"rf_use_mask_{func_name}")
            if config_value is not None:
                return config_value

    if default_false_for_behavior_version_up_to is not None:
        from returnn.util.basic import BehaviorVersion

        if BehaviorVersion.get() <= default_false_for_behavior_version_up_to:
            return False
    return default


def _get_opt_bool_from_config(config: Config, key: str) -> Optional[bool]:
    if key in config.typed_dict:
        config_value = config.typed_dict[key]
        assert config_value is None or isinstance(config_value, bool)
        return config_value
    elif key in config.dict:
        return config.bool(key, None)
    return None



"""
Usage:

Create ~/.pypirc with info:

    [distutils]
    index-servers =
        pypi

    [pypi]
    repository: https://upload.pypi.org/legacy/
    username: ...
    password: ...

(Not needed anymore) Registering the project: python3 setup.py register
New release: python3 setup.py sdist upload

I had some trouble at some point, and this helped:
pip3 install --user twine
python3 setup.py sdist
twine upload dist/*.tar.gz

See also MANIFEST.in for included files.

For debugging this script:

python3 setup.py sdist
pip3 install --user dist/*.tar.gz -v
(Without -v, all stdout/stderr from here will not be shown.)

"""

import os
import shutil
from returnn.__setup__ import get_version_str, debug_print_file


def main():
    """
    Setup main entry
    """
    # Do not use current time as fallback for the version anymore,
    # as this would result in a version which can be bigger than what we actually have,
    # so this would not be useful at all.
    long_version = get_version_str(verbose=True, fallback="1.0.0+setup-fallback-version", long=True)
    version = long_version[: long_version.index("+")]

    if os.environ.get("DEBUG", "") == "1":
        debug_print_file(".")
        debug_print_file("PKG-INFO")
        debug_print_file("pip-egg-info")
        debug_print_file("pip-egg-info/returnn.egg-info")
        debug_print_file("pip-egg-info/returnn.egg-info/SOURCES.txt")  # like MANIFEST

    if os.path.exists("PKG-INFO"):
        if os.path.exists("MANIFEST"):
            print("package_data, found PKG-INFO and MANIFEST")
            package_data = open("MANIFEST").read().splitlines() + ["PKG-INFO"]
        else:
            print("package_data, found PKG-INFO, no MANIFEST, use *")
            # Currently the setup will ignore all other data except in returnn/.
            # At least make the version available.
            shutil.copy("PKG-INFO", "returnn/")
            shutil.copy("_setup_info_generated.py", "returnn/")
            # Just using package_data = ["*"] would only take files from current dir.
            package_data = []
            for root, dirs, files in os.walk("."):
                for file in files:
                    package_data.append(os.path.join(root, file))
    else:
        print("dummy package_data, does not matter, likely you are running sdist")
        with open("_setup_info_generated.py", "w") as f:
            f.write("version = %r\n" % version)
            f.write("long_version = %r\n" % long_version)
        package_data = ["MANIFEST", "_setup_info_generated.py"]

    from distutils.core import setup

    setup(
        name="returnn",
        version=version,
        packages=["returnn"],
        include_package_data=True,
        package_data={"returnn": package_data},  # filtered via MANIFEST.in
        install_requires=[
            "numpy",
        ],
        description="The RWTH extensible training framework for universal recurrent neural networks",
        author="Albert Zeyer",
        author_email="albzey@gmail.com",
        url="https://github.com/rwth-i6/returnn/",
        license="RETURNN license",
        long_description=open("README.rst").read(),
        long_description_content_type="text/x-rst",
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Environment :: GPU",
            "Environment :: GPU :: NVIDIA CUDA",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: Other/Proprietary License",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )


if __name__ == "__main__":
    main()



"""
File cache.

Copies files from a remote filesystem (e.g. NFS)
to a local filesystem (e.g. /var/tmp) to speed up access.

See https://github.com/rwth-i6/returnn/issues/1519 for initial discussion.

Main class is :class:`FileCache`.
"""

from __future__ import annotations
from typing import Any, Collection, Dict, Iterable, List, Optional, Tuple, Union
import errno
import os
import pathlib
import time
import shutil
import dataclasses
from dataclasses import dataclass
from collections import defaultdict
from contextlib import contextmanager
import json
from threading import Lock, Thread, Event
from returnn.config import Config, get_global_config
from .basic import expand_env_vars, LockFile, human_bytes_size


__all__ = ["FileCache", "CachedFile", "get_instance"]


class FileCache:
    """
    File cache.

    Copies files from a remote filesystem (e.g. NFS)
    to a local filesystem (e.g. ``/var/tmp``) to speed up access.

    Some assumptions we depend on:

    - When a cached file is available,
      and its size matches the original file
      and its mtime is not older than the original file,
      we can use it.
    - We will update the cached file mtime frequently (every second) via a background thread
      of used cached files, to mark that they are used.
      (We would maybe want to use atime, but we don't expect that atime can be relied on.)
      Note that updating mtime might influence the behavior of some external tools.
    - :func:`os.utime` will update mtime, and mtime is somewhat accurate (up to 10 secs maybe),
      mtime compares to time.time().
    - :func:`shutil.disk_usage` can be relied on.

    See https://github.com/rwth-i6/returnn/issues/1519 for initial discussion.
    """

    def __init__(
        self,
        *,
        cache_directory: str = "$TMPDIR/$USER/returnn/file_cache",
        cleanup_files_always_older_than_days: float = 31.0,
        cleanup_files_wanted_older_than_days: float = 1.0,
        cleanup_disk_usage_wanted_free_ratio: float = 0.2,  # try to free at least 20% disk space
        cleanup_disk_usage_wanted_multiplier: float = 2.0,  # try to free 2x the required space for a file
        num_tries: int = 3,  # retry twice by default
    ):
        """
        :param cache_directory: directory where to cache files.
            Uses :func:`expand_env_vars` to expand environment variables.
        :param cleanup_files_always_older_than_days: always cleanup files older than this.
        :param cleanup_files_wanted_older_than_days: if cleanup_disk_usage_wanted_free_ratio not reached,
            cleanup files older than this.
        :param cleanup_disk_usage_wanted_free_ratio: try to free at least this ratio of disk space.
        :param cleanup_disk_usage_wanted_multiplier: when making space for a new file, try to free at
            least this times as much space.
        :param num_tries: how many times to try caching a file before giving up
        """
        self.cache_directory = expand_env_vars(cache_directory)
        self._cleanup_files_always_older_than_days = cleanup_files_always_older_than_days
        self._cleanup_files_wanted_older_than_days = cleanup_files_wanted_older_than_days
        self._cleanup_disk_usage_wanted_free_ratio = cleanup_disk_usage_wanted_free_ratio
        assert cleanup_disk_usage_wanted_multiplier >= 1.0
        self._cleanup_disk_usage_wanted_multiplier = cleanup_disk_usage_wanted_multiplier
        self._touch_files_thread = _TouchFilesThread(cache_base_dir=self.cache_directory)
        assert num_tries > 0
        self._num_tries = num_tries

    # Note on lock_timeout: It will check whether a potentially existing lock file is older than this timeout,
    # and if so, then it would delete the existing lock file, assuming it is from a crashed previous run.
    # We are always keeping the lock file mtime updated via the _touch_files_thread (every second),
    # so it should never be older than this timeout.
    # If there is a Python exception anywhere here, we will always properly release the lock.
    # Only if the process dies (e.g. killed, segfault or so), the lock file might be left-over,
    # and another process might need to wait for this timeout.
    # We don't expect that this must be configured, so let's just use a reasonable default.
    # This should be more than the _touch_files_thread interval (1 sec).
    _lock_timeout = 20

    def __del__(self):
        self._touch_files_thread.stop.set()

    def get_file(self, src_filename: str) -> str:
        """
        Get cached file.
        This will copy the file to the cache directory if it is not already there.
        This will also make sure that the file is not removed from the cache directory
        via the _touch_files_thread
        until you call :func:`release_file`.

        :param src_filename: source file to copy (if it is not already in the cache).
        :return: cached file path (in the cache directory)
        """
        dst_filename = self._get_dst_filename(src_filename)
        last_error = None
        for try_nr in range(self._num_tries):
            if try_nr > 0:
                print(
                    f"FileCache: Ignoring error while copying {dst_filename}: {type(last_error).__name__}: {last_error}"
                )
                time.sleep(1)
            try:
                self._copy_file_if_needed(src_filename, dst_filename)
                break
            except OSError as e:
                if e.errno == errno.ENOSPC:
                    last_error = e
                else:
                    raise e
        if last_error is not None:
            raise last_error
        # protect info file from tempreaper, which looks at the mtime
        self._touch_files_thread.files_extend([dst_filename, self._get_info_filename(dst_filename)])
        return dst_filename

    def release_files(self, filenames: Union[str, Iterable[str]]):
        """
        Release cached files.
        This just says that we are not using the files anymore for now.
        They will be kept in the cache directory for now,
        and might be removed when the cache directory is cleaned up.

        :param filenames: files to release (paths in the cache directory)
        """
        if isinstance(filenames, str):
            filenames = [filenames]
        self._touch_files_thread.files_remove(fn_ for fn in filenames for fn_ in [fn, self._get_info_filename(fn)])

    def cleanup(self, *, need_at_least_free_space_size: int = 0) -> CleanupResult:
        """
        Cleanup cache directory.
        """
        if not os.path.exists(self.cache_directory):
            return CleanupResult(abort_reason="cache directory does not exist")
        disk_usage = shutil.disk_usage(self.cache_directory)
        want_free_space_size = max(
            int(self._cleanup_disk_usage_wanted_multiplier * need_at_least_free_space_size),
            int(self._cleanup_disk_usage_wanted_free_ratio * disk_usage.total),
        )
        cleanup_timestamp_file = self.cache_directory + "/.recent_full_cleanup"
        try:
            last_full_cleanup = os.stat(cleanup_timestamp_file).st_mtime
        except FileNotFoundError:
            last_full_cleanup = float("-inf")
        # Get current time now, so that cur_time - mtime is pessimistic,
        # and does not count the time for the cleanup itself.
        cur_time = time.time()
        # If we have enough free space, and we did a full cleanup recently, we don't need to do anything.
        if want_free_space_size <= disk_usage.free and cur_time - last_full_cleanup < 60 * 10:
            return CleanupResult(
                abort_reason=(
                    f"enough free space"
                    f" ({human_bytes_size(want_free_space_size)} < {human_bytes_size(disk_usage.free)})"
                    f" and recent cleanup ({(cur_time - last_full_cleanup) / 60:.1f} minutes ago)"
                )
            )
        # immediately update the file's timestamp to reduce racyness between worker processes
        # Path().touch() also creates the file if it doesn't exist yet
        pathlib.Path(cleanup_timestamp_file).touch(exist_ok=True)
        # Do a full cleanup, i.e. iterate through all files in cache directory and check their mtime.
        all_files = []  # mtime, neg size (better for sorting), filename
        for root, dirs, files in os.walk(self.cache_directory):
            for rel_fn in files:
                fn = root + "/" + rel_fn
                if fn == cleanup_timestamp_file:
                    continue
                elif self._is_info_filename(fn):
                    # skip keepalive files, they are processed together with the file they guard
                    continue
                elif self._is_lock_filename(fn):
                    # skip lock files, removing them would accidentally release locks
                    continue
                try:
                    f_stat = os.stat(fn)
                except Exception as exc:
                    print(f"FileCache: Error while stat {fn}: {type(exc).__name__}: {exc}")
                    continue
                else:
                    all_files.append((f_stat.st_mtime, -f_stat.st_blocks * 512, fn))
        all_files.sort()
        cur_expected_free = disk_usage.free
        reached_more_recent_files = False
        cur_used_time_threshold = self._lock_timeout * 0.5  # Used files mtime should be updated every second.
        total_cache_files_size = sum(-neg_size for _, neg_size, _ in all_files)
        total_cur_used_cache_files_size = sum(
            -neg_size for mtime, neg_size, fn in all_files if cur_time - mtime <= cur_used_time_threshold
        )
        report_size_str = (
            f"Total size cached files: {human_bytes_size(total_cache_files_size)},"
            f" currently used: {human_bytes_size(total_cur_used_cache_files_size)}"
        )
        for mtime, neg_size, fn in all_files:
            size = -neg_size
            delete_reason = None

            lock_dir, lock_file_name = self._get_lock_filename(fn)
            lock_file = LockFile(directory=lock_dir, name=lock_file_name, lock_timeout=self._lock_timeout)
            if not lock_file.try_lock():
                print(f"FileCache: lock for {fn} is currently held, skipping.")
                continue
            try:
                # Re-check mtime with lock, could have been updated by another
                # process in the meantime.
                # We do not update the `mtime` variable here, because the code
                # assumes that the list of files is sorted by mtime to abort
                # early when enough space has been made.
                # Instead, we treat the case where the mtime was updated during
                # cleanup as an outlier and continue as if no other mtimes had
                # changed.
                # See for discussion:
                #   - https://github.com/rwth-i6/returnn/issues/1675
                #   - https://github.com/rwth-i6/returnn/pull/1709
                try:
                    cur_mtime = os.stat(fn).st_mtime
                except FileNotFoundError:
                    # File was deleted while waiting for the lock, or because it was
                    # a temporary copy file and was renamed to its final location.
                    # Since we don't know whether it was actually deleted or just
                    # renamed, we leave cur_expected_free unchanged.
                    continue
                except Exception as exc:
                    print(f"FileCache: Error refreshing mtime of {fn}: {type(exc).__name__}: {exc}")
                    continue
                if cur_mtime > mtime and (time.time() - cur_mtime) <= cur_used_time_threshold:
                    print(f"FileCache: {fn} has been updated during cleanup, skipping.")
                    continue
                if cur_time - mtime > self._cleanup_files_always_older_than_days * 60 * 60 * 24:
                    delete_reason = f"File is {(cur_time - mtime) / 60 / 60 / 24:.1f} days old"
                else:
                    reached_more_recent_files = True
                if not delete_reason and need_at_least_free_space_size > cur_expected_free:
                    # Still must delete some files.
                    if cur_time - mtime > cur_used_time_threshold:
                        delete_reason = f"Still need more space, file is {(cur_time - mtime) / 60 / 60:.1f} hours old"
                    else:
                        raise Exception(
                            f"We cannot free enough space on {self.cache_directory}.\n"
                            f"Needed: {human_bytes_size(need_at_least_free_space_size)},\n"
                            f"currently available: {human_bytes_size(cur_expected_free)},\n"
                            f"oldest file is still too recent: {fn}.\n"
                            f"{report_size_str}"
                        )
                if not delete_reason and want_free_space_size > cur_expected_free:
                    if cur_time - mtime > self._cleanup_files_wanted_older_than_days * 60 * 60 * 24:
                        delete_reason = f"Still want more space, file is {(cur_time - mtime) / 60 / 60:.1f} hours old"
                    else:
                        # All further files are even more recent, so we would neither cleanup them,
                        # so we can also just stop now.
                        break

                if delete_reason:
                    cur_expected_free += size
                    print(
                        f"FileCache: Delete file {fn}, size {human_bytes_size(size)}. {delete_reason}."
                        f" After deletion, have {human_bytes_size(cur_expected_free)} free space."
                    )
                    try:
                        os.remove(fn)
                    except Exception as exc:
                        if not isinstance(exc, FileNotFoundError):
                            print(f"FileCache: Error while removing {fn}: {type(exc).__name__}: {exc}")

                        # We don't know whether the file was just renamed or actually deleted, so
                        # we do as if its space has not been freed.
                        cur_expected_free -= size
                    try:
                        os.remove(self._get_info_filename(fn))
                    except FileNotFoundError:
                        pass
                    except Exception as exc:
                        print(f"FileCache: Ignoring error file removing info file of {fn}: {type(exc).__name__}: {exc}")
            finally:
                lock_file.unlock()

            if reached_more_recent_files and want_free_space_size <= cur_expected_free:
                # Have enough free space now.
                break

        if need_at_least_free_space_size > cur_expected_free:
            raise Exception(
                f"We cannot free enough space on {self.cache_directory}.\n"
                f"Needed: {human_bytes_size(need_at_least_free_space_size)},\n"
                f"currently available: {human_bytes_size(cur_expected_free)}.\n"
                f"{report_size_str}"
            )

        # Cleanup empty dirs.
        for root, dirs, files in os.walk(self.cache_directory, topdown=False):
            if files:
                continue
            try:
                if cur_time - os.stat(root).st_mtime <= cur_used_time_threshold:  # still in use?
                    continue
            except Exception as exc:
                print(f"FileCache: Error while stat dir {root}: {type(exc).__name__}: {exc}")
                continue
            try:
                # Recheck existence of dirs, because they might have been deleted by us.
                if any(os.path.exists(root + "/" + d) for d in dirs):
                    continue
            except Exception as exc:
                print(f"FileCache: Error while checking sub dirs in {root}: {type(exc).__name__}: {exc}")
                continue
            try:
                # We can delete this empty dir.
                print(f"FileCache: Remove empty dir {root}")
                os.rmdir(root)
            except Exception as exc:
                print(f"FileCache: Error while removing empty dir {root}: {type(exc).__name__}: {exc}")

        return CleanupResult(freed=cur_expected_free - disk_usage.free)

    def handle_cached_files_in_config(self, config: Any) -> Tuple[Any, List[str]]:
        """
        :param config: some config, e.g. dict, or any nested structure
        :return: modified config, where all :class:`CachedFile` instances are replaced by the cached file path,
            and the list of cached files which are used.
        """
        import tree

        res_files = []

        def _handle_value(value):
            if isinstance(value, CachedFile):
                res = self.get_file(value.filename)
                res_files.append(res)
                return res
            return value

        return tree.map_structure(_handle_value, config), res_files

    def _get_dst_filename(self, src_filename: str) -> str:
        """
        Get the destination filename in the cache directory.
        """
        assert src_filename.startswith("/")
        dst_file_name = self.cache_directory + src_filename
        return dst_file_name

    @staticmethod
    def _get_info_filename(filename: str) -> str:
        """:return: the name of the corresponding info file to `filename`."""
        return f"{filename}.returnn-info"

    @staticmethod
    def _get_lock_filename(filename: str) -> Tuple[str, str]:
        """:return: lock file target directory and lock file name"""
        return os.path.dirname(filename), os.path.basename(filename) + ".returnn-lock"

    @staticmethod
    def _is_lock_filename(filename: str) -> bool:
        """:return: whether `filename` points to a lock file."""
        return filename.endswith(".returnn-lock")

    @staticmethod
    def _is_info_filename(filename: str) -> bool:
        """:return: whether `filename` points to a info file."""
        return filename.endswith(".returnn-info")

    def _copy_file_if_needed(self, src_filename: str, dst_filename: str):
        """
        Copy the file to the cache directory.
        """
        # Create dirs.
        dst_dir = os.path.dirname(dst_filename)
        os.makedirs(dst_dir, exist_ok=True)

        lock_dir, lock_file = self._get_lock_filename(dst_filename)
        info_file_name = self._get_info_filename(dst_filename)

        # Copy the file, while holding a lock. See comment on lock_timeout above.
        with LockFile(
            directory=lock_dir, name=lock_file, lock_timeout=self._lock_timeout
        ) as lock, self._touch_files_thread.files_added_context(lock.lockfile):
            # Maybe it was copied in the meantime, while waiting for the lock.
            if self._check_existing_copied_file_maybe_cleanup(src_filename, dst_filename):
                print(f"FileCache: using existing file {dst_filename}")
                # Update mtime while holding lock, to synchronize with any concurrent cleanup.
                # See for discussion:
                #   - https://github.com/rwth-i6/returnn/issues/1675
                #   - https://github.com/rwth-i6/returnn/pull/1709
                os.utime(dst_filename, None)
                os.utime(info_file_name, None)
                # Ensure we proactively make space for other users
                # even in case we have all files ready on disk.
                # See for discussion: https://github.com/rwth-i6/returnn/pull/1752.
                self.cleanup(need_at_least_free_space_size=0)
                return

            print(f"FileCache: Copy file {src_filename} to cache")

            # Make sure we have enough disk space, st_size +1 due to _copy_with_prealloc
            self.cleanup(need_at_least_free_space_size=os.stat(src_filename).st_size + 1)

            dst_tmp_filename = dst_filename + ".copy"
            if os.path.exists(dst_tmp_filename):
                # The minimum age should be at least the lock_timeout.
                # (But leave a bit of room for variance in timing in the sanity check below.)
                dst_tmp_file_age = time.time() - os.stat(dst_tmp_filename).st_mtime
                assert dst_tmp_file_age > self._lock_timeout * 0.8, (
                    f"FileCache: Expected left-over temp copy file {dst_tmp_filename}"
                    f" from crashed previous copy attempt"
                    f" to be older than {self._lock_timeout * 0.8:.1f}s but it is {dst_tmp_file_age} seconds old"
                )

            with self._touch_files_thread.files_added_context(dst_dir):
                # save mtime before the copy process to have it pessimistic
                orig_mtime_ns = os.stat(src_filename).st_mtime_ns
                FileInfo(mtime_ns=orig_mtime_ns).save(info_file_name)

                try:
                    _copy_with_prealloc(src_filename, dst_tmp_filename)
                except Exception:
                    # Cleanup if it was created already.
                    # That avoids some of the ambiguity of the existence of the .copy file.
                    # https://github.com/rwth-i6/returnn/issues/1785
                    try:
                        os.remove(dst_tmp_filename)
                    except FileNotFoundError:
                        pass
                    try:
                        os.remove(info_file_name)
                    except FileNotFoundError:  # not really expected here, but safe to ignore
                        pass
                    raise
                os.rename(dst_tmp_filename, dst_filename)

    @staticmethod
    def _check_existing_copied_file_maybe_cleanup(src_filename: str, dst_filename: str) -> bool:
        """
        Check if the file is in the cache directory.
        """
        if not os.path.exists(dst_filename):
            return False
        src_stat = os.stat(src_filename)
        dst_stat = os.stat(dst_filename)
        try:
            last_known_mtime_ns = FileInfo.load(FileCache._get_info_filename(dst_filename)).mtime_ns
        except FileNotFoundError:
            # for existing setups where files may be cached, but the info file has
            # not been written to disk yet, we re-fetch the file
            last_known_mtime_ns = None
        if (
            src_stat.st_size != dst_stat.st_size
            or last_known_mtime_ns is None
            or src_stat.st_mtime_ns > last_known_mtime_ns
        ):
            os.remove(dst_filename)
            return False
        return True


class CleanupResult:
    """
    Result from :func:`FileCache.cleanup`.
    """

    def __init__(self, *, freed: int = 0, abort_reason: Optional[str] = None):
        self.freed = freed
        self.abort_reason = abort_reason

    def __repr__(self):
        if self.abort_reason:
            return f"CleanupResult(abort_reason={self.abort_reason})"
        return f"CleanupResult(freed={human_bytes_size(self.freed)})"


def get_instance(config: Optional[Config] = None) -> FileCache:
    """
    Returns a file cache instance potentially initialized by the global config.

    Uses defaults if no global config is set.
    """
    config = config or get_global_config(return_empty_if_none=True)
    kwargs = config.typed_value("file_cache_opts") or {}
    return FileCache(**kwargs)


def _copy_with_prealloc(src: str, dst: str):
    """
    Copies the file at `src` to `dst` preallocating the space at `dst` before the
    copy to reduce the chance of race conditions w/ free-disk-space checks occuring.

    Note the function preallocates `size + 1` to allow detecting incompletely copied
    files by a mismatch in the file size, should the copy process be interrupted. The
    additional byte is then truncated away after copying.

    In practice this function is used to copy to a temporary file first, so the
    +1-size trick is technically not necessary -- but it also does not hurt leaving
    it in.
    """
    file_size = os.stat(src).st_size
    with open(dst, "wb") as dst_file:
        if file_size > 0:
            # Prealloc size + 1, see docstring for why.
            #
            # See also `_check_existing_copied_file_maybe_cleanup`.
            try:
                os.posix_fallocate(dst_file.fileno(), 0, file_size + 1)
            except AttributeError:
                # posix_fallocate not available (non-Linux unixes), fallback to seek+write
                dst_file.seek(file_size)
                dst_file.write(b"\0")
                dst_file.seek(0)
        with open(src, "rb") as src_file:
            try:
                os.posix_fadvise(src_file.fileno(), 0, file_size, os.POSIX_FADV_SEQUENTIAL)
                os.posix_fadvise(dst_file.fileno(), 0, file_size, os.POSIX_FADV_SEQUENTIAL)
            except AttributeError:
                # posix_fadvise not available (non-Linux unixes)
                pass
            shutil.copyfileobj(src_file, dst_file)
        dst_file.truncate(file_size)


@dataclass
class CachedFile:
    """
    Represents some file to be cached in a user config.
    See :func:`FileCache.handle_cached_files_in_config`.
    """

    filename: str  # original filename


@dataclass(frozen=True)
class FileInfo:
    """
    Represents meta information about cached files.

    We currently save the last known mtime of the source file.
    """

    mtime_ns: int

    def save(self, filename: str):
        """Saves this instance to `filename`."""
        with open(filename, "wt") as f:
            f.write(json.dumps(dataclasses.asdict(self)))

    @classmethod
    def load(cls, filename: str) -> "FileInfo":
        """Loads a previously saved FileInfo from the file at `filename`."""
        with open(filename, "rt") as f:
            data = json.load(f)
        return cls(**data)


class _TouchFilesThread(Thread):
    def __init__(self, *, interval: float = 1.0, cache_base_dir: str):
        super().__init__(daemon=True)
        self.stop = Event()
        self.files = defaultdict(int)  # usage counter
        self.files_lock = Lock()  # lock for self.files/self.locks
        self.locks: Dict[str, Lock] = {}  # filename -> lock
        self.interval = interval
        self.cache_base_dir = cache_base_dir
        self._is_started = False  # careful: `_started` is already a member of the base class

    def run(self):
        """thread main loop"""
        while True:
            # locks dict copied under GIL
            locks = self.locks.copy()
            for filename, lock in locks.items():
                with lock:
                    if filename not in self.files:
                        continue
                    try:
                        os.utime(filename, None)
                    except Exception as exc:
                        print(f"FileCache: failed updating mtime of {filename}: {exc}")
                        raise
            if self.stop.wait(self.interval):
                return

    def start_once(self):
        """reentrant variant of start() that can safely be called multiple times"""
        if self._is_started:
            return
        self._is_started = True
        self.start()

    def files_extend(self, to_add: Union[str, Iterable[str]]):
        """append"""
        if isinstance(to_add, str):
            to_add = [to_add]
        assert isinstance(to_add, Iterable)
        self.start_once()
        # we track the parent directories as well and give them their own locks to be
        # able to synchronize their deletion with the touch thread
        files_to_iter = _files_with_parents(to_add, base_dir=self.cache_base_dir)
        with self.files_lock:  # to avoid that we end up with duplicate locks
            for file, count in files_to_iter.items():
                self.files[file] += count
                if file not in self.locks:
                    self.locks[file] = Lock()

    def files_remove(self, to_remove: Union[str, Iterable[str]]):
        """remove"""
        if isinstance(to_remove, str):
            to_remove = [to_remove]
        assert isinstance(to_remove, Iterable)
        for file, count in _files_with_parents(to_remove, base_dir=self.cache_base_dir).items():
            with self.locks[file], self.files_lock:
                self.files[file] -= count
                if self.files[file] <= 0:
                    del self.files[file]
                    del self.locks[file]

    @contextmanager
    def files_added_context(self, files: Collection[str]):
        """temporarily add files, and remove them afterwards again."""
        self.files_extend(files)
        try:
            yield
        finally:
            self.files_remove(files)


def _all_parent_dirs(filename: str, *, base_dir: str) -> List[str]:
    assert filename.startswith(base_dir + "/")
    dirs = []
    while True:
        filename = os.path.dirname(filename)
        if filename == base_dir:
            break
        dirs.append(filename)
    return dirs


def _files_with_parents(filenames: Iterable[str], *, base_dir: str) -> Dict[str, int]:
    res = defaultdict(int)  # dict to have order deterministic
    for fn in filenames:
        res[fn] += 1
        for fn_ in _all_parent_dirs(fn, base_dir=base_dir):
            res[fn_] += 1
    return res
